package org.xbib.time.pretty;

import static org.junit.jupiter.api.Assertions.assertEquals;
import static org.junit.jupiter.api.Assertions.assertTrue;
import org.junit.jupiter.api.AfterEach;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Test;
import java.time.Instant;
import java.time.LocalDateTime;
import java.time.ZoneId;
import java.util.Locale;

public class PrettyTimeI18n_Test {

    private Locale locale;

    @BeforeEach
    public void setUp() {
        locale = Locale.getDefault();
    }

    @Test
    public void testPrettyTimeDefault() {
        // The default resource bundle should be used
        PrettyTime p = new PrettyTime(0, Locale.ROOT);
        LocalDateTime localDateTime = LocalDateTime.ofInstant(Instant.ofEpochMilli(1), ZoneId.systemDefault());
        assertEquals("moments from now", p.format(localDateTime));
    }

    @Test
    public void testPrettyTimeGerman() {
        // The German resource bundle should be used
        PrettyTime p = new PrettyTime(Locale.GERMAN);
        LocalDateTime ref = LocalDateTime.ofInstant(Instant.ofEpochMilli(0), ZoneId.systemDefault());
        p.setReference(ref);
        LocalDateTime localDateTime = LocalDateTime.ofInstant(Instant.ofEpochMilli(1), ZoneId.systemDefault());
        assertEquals("Jetzt", p.format(localDateTime));
    }

    @Test
    public void testPrettyTimeSpanish() {
        // The Spanish resource bundle should be used
        PrettyTime p = new PrettyTime(new Locale("es"));
        assertEquals("en un instante", p.format(LocalDateTime.now()));
    }

    @Test
    public void testPrettyTimeDefaultCenturies() {
        // The default resource bundle should be used
        PrettyTime p = new PrettyTime((3155692597470L * 3L), Locale.ROOT);
        LocalDateTime localDateTime = LocalDateTime.ofInstant(Instant.ofEpochMilli(0), ZoneId.systemDefault());
        assertEquals("3 centuries ago", p.format(localDateTime));
    }

    @Test
    public void testPrettyTimeGermanCenturies() {
        // The default resource bundle should be used
        PrettyTime p = new PrettyTime((3155692597470L * 3L), Locale.GERMAN);
        LocalDateTime localDateTime = LocalDateTime.ofInstant(Instant.ofEpochMilli(0), ZoneId.systemDefault());
        assertEquals("vor 3 Jahrhunderten", p.format(localDateTime));
    }

    @Test
    public void testPrettyTimeViaDefaultLocaleDefault() {
        // The default resource bundle should be used
        Locale.setDefault(Locale.ROOT);
        PrettyTime p = new PrettyTime((0));
        LocalDateTime localDateTime = LocalDateTime.ofInstant(Instant.ofEpochMilli(1), ZoneId.systemDefault());
        assertEquals("moments from now", p.format(localDateTime));
    }

    @Test
    public void testPrettyTimeViaDefaultLocaleGerman() {
        // The German resource bundle should be used
        Locale.setDefault(Locale.GERMAN);
        PrettyTime p = new PrettyTime((0));
        LocalDateTime localDateTime = LocalDateTime.ofInstant(Instant.ofEpochMilli(1), ZoneId.systemDefault());
        assertEquals("Jetzt", p.format(localDateTime));
    }

    @Test
    public void testPrettyTimeViaDefaultLocaleDefaultCenturies() {
        // The default resource bundle should be used
        Locale.setDefault(Locale.ROOT);
        PrettyTime p = new PrettyTime((3155692597470L * 3L));
        LocalDateTime localDateTime = LocalDateTime.ofInstant(Instant.ofEpochMilli(0), ZoneId.systemDefault());
        assertEquals("3 centuries ago", p.format(localDateTime));
    }

    @Test
    public void testPrettyTimeViaDefaultLocaleGermanCenturies() {
        // The default resource bundle should be used
        Locale.setDefault(Locale.GERMAN);
        PrettyTime p = new PrettyTime((3155692597470L * 3L));
        LocalDateTime localDateTime = LocalDateTime.ofInstant(Instant.ofEpochMilli(0), ZoneId.systemDefault());
        assertEquals("vor 3 Jahrhunderten", p.format(localDateTime));
    }

    @Test
    public void testPrettyTimeRootLocale() {
        long t = 1L;
        PrettyTime p = new PrettyTime(0, Locale.ROOT);
        LocalDateTime localDateTime = LocalDateTime.ofInstant(Instant.ofEpochMilli(0), ZoneId.systemDefault());
        while (1000L * 60L * 60L * 24L * 365L * 1000000L > t) {
            assertTrue(p.format(localDateTime).endsWith("now"));
            t *= 2L;
        }
    }

    @Test
    public void testPrettyTimeGermanLocale() {
        long t = 1L;
        PrettyTime p = new PrettyTime(0, Locale.GERMAN);
        LocalDateTime localDateTime = LocalDateTime.ofInstant(Instant.ofEpochMilli(0), ZoneId.systemDefault());
        while (1000L * 60L * 60L * 24L * 365L * 1000000L > t) {
            assertTrue(p.format(localDateTime).startsWith("in") || p.format(localDateTime).startsWith("Jetzt"));
            t *= 2L;
        }
    }

    @AfterEach
    public void tearDown() {
        Locale.setDefault(locale);
    }

}
